import torch
import torch.nn as nn
from aiia import AIIA, AIIAConfig, AIIABase
from config import UpsamplerConfig

# Upsampler model that uses the configuration from the base model.
class Upsampler(AIIA):
    def __init__(self, base_model: AIIABase):
        # Assume that base_model.config is an instance of UpsamplerConfig.
        super().__init__(base_model.config)
        self.base_model = base_model
        self.config = UpsamplerConfig(kwargs=self.base_model.config)
        # Create the upsample layer using values from the configuration.
        print(self.config.upsample_scale)
        self.upsample = nn.Upsample(
            scale_factor=self.config.upsample_scale,
            mode=self.config.upsample_mode,
            align_corners=self.config.upsample_align_corners
        )

    def forward(self, x):
        x = self.base_model(x)
        x = self.upsample(x)
        return x

    @classmethod
    def load(cls, path: str):
        """
        Load the model:
          - First, load the base model (including its configuration and state_dict).
          - Then, wrap it with the Upsampler class.
          - Finally, load the combined state dictionary.
        """
        base_model = AIIABase.load(path)
        instance = cls(base_model)
        
        device = 'cuda' if torch.cuda.is_available() else 'cpu'
        state_dict = torch.load(f"{path}/model.pth", map_location=device)
        instance.load_state_dict(state_dict)
        return instance


if __name__ == "main":
    from aiia import AIIABase, AIIAConfig
    # Create a configuration and build a base model.
    config = AIIAConfig()
    base_model = AIIABase("test2")
    # Instantiate Upsampler from the base model (works correctly).
    upsampler = Upsampler(base_model)

    # Save the model (both configuration and weights).
    upsampler.save("test2")

    # Now load using the overridden load method; this will load the complete model.
    upsampler_loaded = Upsampler.load("test2")
    print("Updated configuration:", upsampler_loaded.config.__dict__)
